# -*- coding: utf-8 -*-
"""
Created on Tue Mar 24 21:13:51 2020

@author: Jean
"""

from graphviz import Source
from igraph import *
import random
import numpy as np
import pandas as pd
import time
import math

# 读入纳税人交易网络
vdf = pd.read_csv("D:/temp/data/增值税自开票2017/纳税人id1000w.csv")
edf = pd.read_csv("D:/temp//data/增值税自开票2017/纳税人开票汇总1000w.csv")
edf_all = pd.read_csv("D:/temp//data/增值税自开票2017/纳税人开票汇总1000w_alledges.csv")
edf_all.rename(columns={"xfsbh":"source","gfsbh":"target"},inplace=True)

# turn into dictionary and build a graph with Graph.DictList()
vdic = vdf.to_dict(orient="records")
edic = edf.to_dict(orient="records")
g = Graph.DictList(vdic,edic, vertex_name_attr="id", directed=True)

# 读入纳税人识别号对照表及税务登记信息
types = {"ID":int,"NSRSBHMD5":str,"NSRSBH":str,"DJXH":str,"SHXYDM":str,"SSDABH":str,
         "NSRMC":str,"KZZTDJLX_DM":str,"DJZCLX_DM":str,"HY_DM":str,
         "ZGSWSKFJ_DM":str,"NSRZT_DM":str}
nsrs = pd.read_csv("D:/temp//data/增值税自开票2017/纳税人识别号对照表.csv", sep="~", dtype=types)

types = {"spbm":str,"spmc":str}
spbm = pd.read_excel("D:/temp//data/增值税自开票2017/2019年4月税收分类编码Excel版2.xlsx",dtype=types)

# 读入发票明细,3142万行，65秒读入，对于大数据，指定数据类型后读入速度快很多
# 否则当其中一列判断为有混合类型时，速度会慢一倍，iterator=True, 分段读入
t1 = time.time()
types = {"xfid":int,"gfid":int,"hwmc":str,"ggxh":str,"dw":str,"je":float,"spbm":str}
reader = pd.read_csv('D:/temp//data/增值税自开票2017/自开票明细.csv', \
                     iterator=True, sep='~',dtype=types, encoding="utf-8")
try:
    zkfp = reader.get_chunk(100000000)
except StopIteration:
    print ("Iteration is stopped.")
t2 = time.time()
print(t2-t1)


# breadth first search of paths and unique loops
def get_loops(adj, paths, maxlen):
    # tracking the actual path length:
    maxlen -= 1
    nxt_paths = []
    # iterating over all paths:
    for path in paths['paths']:
        # iterating neighbors of the last vertex in the path:
        for nxt in adj[path[-1]]:
            # attaching the next vertex to the path:
            nxt_path = path + [nxt]
            if path[0] == nxt and min(path) == nxt:
                # the next vertex is the starting vertex, we found a loop
                # we keep the loop only if the starting vertex has the 
                # lowest vertex id, to avoid having the same loops 
                # more than once
                paths['loops'].append(nxt_path)
                # if you don't need the starting vertex 
                # included at the end:
                # paths['loops'].append(path)
            elif nxt not in path:
                # keep the path only if we don't create 
                # an internal loop in the path
                nxt_paths.append(nxt_path)
    # paths grown by one step:
    paths['paths'] = nxt_paths
    if maxlen == 0:
        # the final return when maximum search length reached
        return paths
    else:
        # recursive return, to grow paths further
        return get_loops(adj, paths, maxlen)

adj = []
loops = []
# the maximum length to limit computation time on large graphs
# maximum could be vcount(graph), but that might take for ages
maxlen = 4
# creating an adjacency list
# for directed graphs use the 'mode' argument of neighbors() 
# according to your needs ('in', 'out' or 'all')
t1 = time.time()
adj = [[n.index for n in v.neighbors("out")] for v in g.vs]
t2 = time.time()
print("time: ",t2-t1)

# recursive search of loops 
# for each vertex as candidate starting point
t1 = time.time()
loops = []
for start in range(g.vcount()):
    lps = get_loops(adj, 
        {'paths': [[start]], 'loops': []}, maxlen)['loops']
    loops += lps
    t2 = time.time()
    print("vertex: ",start," loops:",len(lps),"time: ",int(t2-t1))

print(len(loops))


# 发现流动金额稳定的有向环，波动的阈值为percent
def stable_loops(g,loops,percent=0.9):
    t1 = time.time()
    slps = []
    l=0
    for path in loops:
        pl = len(path); mx = None; mn = None; sm = 0
        je = list()
        for i in range(pl-1):
            source = path[i]; target = path[i+1]
            e = g.es[g.get_eid(source, target)]
            je.append(e["je"])
            if i==0:
                mx = e["je"]; mn = e["je"]; sm = e["je"]
            else:
                sm = e["je"] + sm
                if e["je"]>mx:
                    mx =e["je"]
                if e["je"]< mn:
                    mn = e["je"]
        mean = sm/(pl-1)
        if mn/mean < percent or mx/mean > 2-percent:
            isStable = False
        else:
            isStable = True
            slps =slps + [path]
            
        l +=1
        t2 = time.time()
        print(l,path, isStable, int(t2-t1))
    
    return slps
                
# 检测波动不超过10%的有向环        
sloops = stable_loops(g,loops,percent=0.8)
print(len(sloops)) 
sloops[0]           
        

# sum for vertex for in&out degrees of cycles, total in&out money amount
# list is much more faster than pandas data frame
def sum_loops(g,sloops):
    t1 = time.time()
    ids = set()
    for path in sloops:
        ids = ids|set(path)
    ids = list(ids)
    ids.sort()
    ind = [0]*len(ids); inje = ind.copy(); outd = ind.copy(); outje = ind.copy()
    # deal with every single cycle
    for path in sloops:
        pl = len(path)
        # deal every edge in the cycle
        for i in range(pl-1):
            source = path[i]; target = path[i+1]
            si = ids.index(source); ti = ids.index(target)
            e = g.es[g.get_eid(source, target)]
            outd[si] =outd[si]+1; outje[si] = outje[si]+e["je"]
            ind[ti] =ind[ti]+1; inje[ti] = inje[ti]+e["je"]            
        t2 = time.time()
        print(path,int(t2-t1))
 
    # turn to a pandas data frame
    idsum = pd.DataFrame({"id":ids,"ind":ind,"inje":inje,"outd":outd,"outje":outje})
    return idsum

idsum = sum_loops(g,sloops)
idIn = idsum.groupby("ind")["ind"].count()
idsum2 = idsum[idsum.ind>=10]

# return subgraph of cycles that contains vertex v
def cyclesub(g,loops,v):
    subset = set()
    for loop in loops:
        loopset = set(loop)
        if v in loopset:
            subset = subset | loopset
    return g.subgraph(subset,implementation="create_from_scratch")

vertex = 1843
# vertex = 2940
g1 = cyclesub(g,sloops,vertex)
g1.vs["label"] = [str(v["id"])+" : "+str(round(v["xje"]/1000000,2)) for v in g1.vs]
#g1.vs["size"] = [round(v["xje"]/100000000,2) for v in g1.vs]
g1.vs["label_dist"]=1.5
g1.es["label"]=[" "+str(round(e["je"]/1000000,2))+" " for e in g1.es]
g1.es["weight"]=g1.es["label"]
#g1.es["width"]=g1.es["label"]
plot(g1,bbox=(800,600))

# who is it?
nsrs[nsrs.NSRSBHMD5==g.vs[vertex]["nsrsbh"]].NSRMC

# check detail for one loop
sloopsv = []
for loop in sloops:
    if vertex in loop:
        sloopsv.append(loop)
print(len(sloopsv))  
sloopsv  

# 查看其中一条环路
loop = sloopsv[1]
g2 = g.subgraph(loop,implementation="create_from_scratch")
g2.vs["label"] = [str(v["id"])+" : "+str(round(v["xje"]/10000,2)) for v in g2.vs]
#g1.vs["size"] = [round(v["xje"]/100000000,2) for v in g1.vs]
g2.vs["label_dist"]=1.5
g2.es["label"]=[" "+str(round(e["je"]/10000,2))+" " for e in g2.es]
g2.es["color"] = "black"
for i in range(len(loop)-1):
    source = g2.vs.select(id_eq=loop[i]); target = g2.vs.select(id_eq=loop[i+1])
    es = g2.es.select(_between=(source, target))
    for edge in es:
        if edge.source == source[0].index:
            edge["color"] = "darkgreen"
        else:
            edge["color"] = "red"
plot(g2,bbox=(500,300))

# 用Graphviz画图，可以正确的显示
# 设置结点与边标签的字体，否则会显示为乱码
g2.vs["fontname"] = "KaiTi"
g2.es["fontname"] = "KaiTi"
# 中间用一个Graphviz格式数据文件连接，马上删除
g2.write_dot("zkp.gv")
with open("zkp.gv","r", encoding ='utf-8') as f:
    dot_graph = f.read()
import os
os.remove("zkp.gv")
from graphviz import Source
# 输出成png格式, layout： engine="dot", 还有 neato, fdp, sfdp,twopi,circo
gv = Source(dot_graph, format="png",engine="dot")
# 打开照片查看器查看，在照片查看器中在打开画图查看，另存成png图
gv.view()

# 输出该闭环的纳税人识别号
g.vs[sloopsv[-1]]["nsrsbh"]

def loop_details(g,edf_all,loop,unit=1,label_name=True):
    nsrsbhs = []
    for vertex in g.vs[loop]:        
        nsrsbhs.append(vertex["nsrsbh"])
    nsrsbhs = pd.DataFrame({"nsrsbh":nsrsbhs})
    nsrsbhs.drop_duplicates(subset = ['nsrsbh'],keep='first',inplace=True)
    nsrsbhs =  pd.merge(nsrsbhs,nsrs,left_on = "nsrsbh", right_on="NSRSBHMD5") 
    nsrsbhs.drop_duplicates(subset = ['NSRSBHMD5','NSRMC'],keep='first',inplace=True)
    nsrsbhs.drop_duplicates(subset = ['NSRSBHMD5'],keep='first',inplace=True)
    nsrsbhs.drop(columns=["nsrsbh"],inplace=True)
    vertices = list()
    for nsrsbh in list(nsrsbhs['NSRSBHMD5']):
        vertex = g.vs.select(nsrsbh_eq=nsrsbh)[0]
        vertices.append(vertex.attributes())    
    edges = edf_all[edf_all.source.isin(g.vs[loop]["nsrsbh"]) &  \
                    edf_all.target.isin(g.vs[loop]["nsrsbh"])]
    edges = pd.merge(edges,spbm,how="left",on="spbm")
    edges_dic = edges.to_dict(orient="records")
    g3 = Graph.DictList(vertices,edges_dic, vertex_name_attr='nsrsbh',  directed=True)
    g3.vs["nsrmc"] = list(nsrsbhs["NSRMC"])
    if label_name:
        g3.vs["label"] = [v["nsrmc"]+" : "+str(round(v["xje"]/unit,2)) for v in g3.vs]
    else:
        g3.vs["label"] = [str(v["id"])+" : "+str(round(v["xje"]/unit,2)) for v in g3.vs]
    g3.es["label"]=[e["spmc"]+str(round(e["je"]/unit,2))+" " for e in g3.es]
    g3.es["color"] = "black"   
    for i in range(len(loop)-1):
        source = g3.vs.select(id_eq=loop[i]); target = g3.vs.select(id_eq=loop[i+1])
        es = g3.es.select(_between=(source, target))
        for edge in es:
            if edge.source == source[0].index:
                edge["color"] = "darkgreen"
            else:
                edge["color"] = "red"
    return g3,nsrsbhs

# 查看其中一条闭环的交易详情，返回该闭环的子图，各结点纳税人的登记资料
# label_name=False输出结点id作标签,金额单位是10000
g3,nsrsl = loop_details(g,edf_all,sloopsv[1],unit=10000,label_name=False)
summary(g3)
plot(g3,bbox=(500,300))
# label_name=True输出纳税人名称作标签
# g3,nsrsl = loop_details(g,edf_all,sloopsv[1],unit=10000,label_name=True)
# g3.vs["size"] = [round(v["xje"]/100000000,2) for v in g3.vs]
# g3.vs["label_dist"]=1.5
# g3.es["weight"]=g3.es["label"]
# plot(g3,bbox=(500,300))

# 用Graphviz画图，可以正确的显示
# 设置结点与边标签的字体，否则会显示为乱码
g3.vs["fontname"] = "KaiTi"
g3.es["fontname"] = "KaiTi"
# 中间用一个Graphviz格式数据文件连接，马上删除
g3.write_dot("zkp.gv")
with open("zkp.gv","r", encoding ='utf-8') as f:
    dot_graph = f.read()
import os
os.remove("zkp.gv")
from graphviz import Source
# 输出成png格式, layout： engine="dot", 还有 neato, fdp, sfdp,twopi,circo
gv = Source(dot_graph, format="png",engine="dot")
# 打开照片查看器查看，在照片查看器中在打开画图查看，另存成png图
gv.view()


def loop_details2(g,zkfp,loop,unit=1,label_name=True):
    nsrsbhs = []
    for vertex in g.vs[loop]:        
        nsrsbhs.append(vertex["nsrsbh"])
    nsrsbhs = pd.DataFrame({"nsrsbh":nsrsbhs})
    nsrsbhs.drop_duplicates(subset = ['nsrsbh'],keep='first',inplace=True)
    nsrsbhs =  pd.merge(nsrsbhs,nsrs,left_on = "nsrsbh", right_on="NSRSBHMD5") 
    nsrsbhs.drop_duplicates(subset = ['NSRSBHMD5','NSRMC'],keep='first',inplace=True)
    nsrsbhs.drop_duplicates(subset = ['NSRSBHMD5'],keep='first',inplace=True)
    nsrsbhs.drop(columns=["nsrsbh"],inplace=True)
    nsrsbhs.reset_index(drop=True)
    vertices = list()
    ids = list(nsrsbhs["ID"]);    i = 0
    for nsrsbh in list(nsrsbhs['NSRSBHMD5']):
        vertex = g.vs.select(nsrsbh_eq=nsrsbh)[0]
        attrs  = vertex.attributes()
        attrs["id"] = ids[i]
        vertices.append(attrs);   i = i+1
    edges = zkfp[zkfp.xfid.isin(list(nsrsbhs["ID"])) & zkfp.gfid.isin(list(nsrsbhs["ID"]))]
    edges = edges[edges.xfid!=edges.gfid]
    edges1 = pd.pivot_table(edges, index=["xfid","gfid","spbm","hwmc"],\
                            values=['je'],aggfunc=np.sum).reset_index()
    edges2 = pd.pivot_table(edges, index=["xfid","gfid","spbm","hwmc"],\
                            values=['je'],aggfunc=len).reset_index()
    edges2.rename(columns={"je":"c"},inplace=True)
    edgest = pd.merge(edges2,edges1,on=["xfid","gfid","spbm","hwmc"])
    edgest = pd.merge(edgest,spbm,how="left",on="spbm")
    edgest.rename(columns={"xfid":"source","gfid":"target"},inplace=True)
    edges_dic = edgest.to_dict(orient="records")
    g3 = Graph.DictList(vertices,edges_dic, vertex_name_attr='id',  directed=True)
    g3.vs["nsrmc"] = list(nsrsbhs["NSRMC"])
    if label_name:
        g3.vs["label"] = [v["nsrmc"]+" : "+str(round(v["xje"]/unit,2)) for v in g3.vs]
    else:
        g3.vs["label"] = [str(v["id"])+" : "+str(round(v["xje"]/unit,2)) for v in g3.vs]
    g3.es["label"]=[e["hwmc"]+str(round(e["je"]/unit,2))+" " for e in g3.es]
    g3.es["color"] = "black"
    ids =pd.merge(pd.DataFrame({"NSRSBHMD5":g.vs[loop]["nsrsbh"]}),\
             nsrsbhs[["ID","NSRSBHMD5"]],how="left")["ID"]
    for i in range(len(ids)-1):
        source = g3.vs.select(id_eq=ids[i]); target = g3.vs.select(id_eq=ids[i+1])
        es = g3.es.select(_between=(source, target))
        for edge in es:
            if edge.source == source[0].index:
                edge["color"] = "darkgreen"
            else:
                edge["color"] = "red"
    return g3,nsrsbhs,edges,edgest

# 查看其中一条闭环的发票详情，返回该闭环的子图，各结点纳税人的登记资料,发票明细及按货劳名称的汇总
# label_name=False输出结点id作标签, 因发票明细条数太多，不适合画图，改按货劳名称汇总后画图
t1 = time.time()
g4,nsrsl4,fpmx,fpmx2 = loop_details2(g,zkfp,sloopsv[1],unit=10000,label_name=False)
t2 = time.time()
print(t2-t1)    # 向下钻取环路的发票明细，1~3秒，速度很快了！
summary(g4)
# 用Graphviz画图，可以正确的显示，设置结点与边标签的字体，否则会显示为乱码
g4.vs["fontname"] = "KaiTi"; g4.es["fontname"] = "KaiTi"
g4.write_dot("zkp.gv")
with open("zkp.gv","r", encoding ='utf-8') as f:
    dot_graph = f.read()
import os
os.remove("zkp.gv")
from graphviz import Source
# 输出成png格式, layout： engine="dot", 还有 neato, fdp, sfdp,twopi,circo
gv = Source(dot_graph, format="png",engine="dot")
gv.view()



# pandas dataframe性能测试, 3142万*7列发票明细数据按 销方id+购方id+分类代码 汇总
# 速度与 R语言 dplyr包差不多，1分钟内都是性能非常好的。
t1 = time.time() #汇总开票金额
zkfp1 = pd.pivot_table(zkfp, index=["xfid","gfid","spbm"],\
                        values=['je'],aggfunc=np.sum).reset_index()
t2 = time.time()
print(t2-t1)    # 24秒
t1 = time.time() #汇总明细条数 
zkfp2 = pd.pivot_table(zkfp, index=["xfid","gfid","spbm"],\
                        values=['je'],aggfunc=len).reset_index()
t2 = time.time()
print(t2-t1)    # 41秒

t1 = time.time() #合并汇总结果
zkfp3 = pd.merge(zkfp1,zkfp2,on=["xfid","gfid","spbm"])
t2 = time.time()
print(t2-t1)    # 3秒

import psutil
import os
info = psutil.virtual_memory()
print (u'内存使用：',psutil.Process(os.getpid()).memory_info().rss)
print (u'总内存：',info.total)
print (u'内存占比：',info.percent)
print (u'cpu个数：',psutil.cpu_count())




# transform an igraph network to Orange data tables of vertices and edges
# the "id" attribute identifys a node, and "source" and "target" attributes for an edge
def graph2tables(g,reindex=True,name="network"):
    import pandas as pd
    from Orange.data.pandas_compat import table_from_frame
    # vertices data table
    vattrs=g.vs.attribute_names()
    vertices= pd.DataFrame()
    for vattr in vattrs:
        vertices[vattr] = g.vs[vattr]
    # when created with Graph.DictList(), the id,source,target attributes are strings
    if reindex:
        vertices["id"] = [vertex.index for vertex in g.vs]  #update a vertex's id to it's index
    vtable = table_from_frame(vertices); vtable.X[:,0].astype(np.int)
    vtable.name = name+"*vertices"
    # edges data table
    source=[]; target=[]
    for edge in g.es:
        source.append(edge.tuple[0])    # index of source vertex
        target.append(edge.tuple[1])    # index of target vertex   
    edges = pd.DataFrame({"source":source,"target":target})
    eattrs = g.es.attribute_names()
    try:    # if an igraph has the "name" attribute, it's built with Graph.DistList()
        vertex_names = g.vs["name"] # there'll be extra "source" and "target" attributes
        eattrs.remove("source")     # then need to get rid or it
        eattrs.remove("target")     # since we'll get it from edge.tuple instead
    except:                         # otherwise, the graph is built with Graph.TupleList()
        pass                        # will get KeyError or ValueError, just pass
    for eattr in eattrs:
        edges[eattr] = g.es[eattr]
    if g.is_directed():
        edges["isDirected"] = 1 # True
    else:
        edges["isDirected"] = 0 # False
    etable=table_from_frame(edges);etable.X[:,0].astype(np.int);etable.X[:,1].astype(np.int)
    etable.name = name+"*edges"
    # return two tables of vertices and edges
    return(vtable,etable)


# depth of 6, test for a single vertex, 97s
vertex = 189
loops2 = []
t1 = time.time()
loops2 += get_loops(adj, 
    {'paths': [[vertex]], 'loops': []}, 6)['loops']
t2 = time.time()
print(len(loops2),"time: ",int(t2-t1))
# 检测波动不超过10%的有向环        
sloops2 = stable_loops(g,loops2,percent=0.7)
print(len(sloops2)) 
sloops2[0]  
g5 = cyclesub(g,sloops2,vertex)
g5.vs["label"] = [str(v["id"])+":"+str(v["xje"]) for v in g5.vs]
g5.es["label"]=g5.es["je"]
g5.es["weight"]=g5.es["je"]
# test orange network with directed network
vstb,estb = graph2tables(g5, reindex=False, name="zkpCycle:"+str(vertex))
import sys
import Orange
from Orange.canvas import __main__ as OrangeMain
sys.exit(OrangeMain.main(datasets = [vstb, estb]))


# 保存工作区
from spyder_kernels.utils.iofuncs import save_dictionary
save_dictionary(globals(),"zkpNetwork.spydata")
# 装入工作区
from spyder_kernels.utils.iofuncs import load_dictionary
globals().update(load_dictionary("zkpNetwork.spydata")[0])
